# -*- coding: utf-8 -*-
'''

 MODULE:       hru-delin_basins

 AUTHOR(S):    adapted from GRASS-HRU (ILMS) - JENA University
               by IRSTEA - Christine Barachet 
               

 PURPOSE:      overlay of all selected layers
			   construction of HRUs

'''

import subprocess, string, os, sys, glob, types, time
import math
import ConfigParser
import grass.script as grass

from osgeo import gdal
from osgeo import ogr

def AWK_cmd_1(cells):
    return "'BEGIN {} {if ($2<=%d) {print $1 \" = \" $1}} END {}'" % cells

def AWK_cmd_2(cells):
    return "'BEGIN {c=0} {if ($2<=%d) {c+=1}} END {print c}'" % cells

def AWK_cmd_3(maxsize):
    return "'BEGIN {old=0} {if (old!=$2&&$2<=%d) {print $2;old=$2}} END {}'" % maxsize

def reclass(map_in, map_out, size):
    os.system("r.stats -lnNc %s | awk %s | r.reclass --o input=%s output=%s" % (map_in, AWK_cmd_1(size), map_in, map_out))

def count_only(map_in, size):
    return int(os.popen("r.stats -lnNc %s | awk %s" % (map_in, AWK_cmd_2(size))).readlines()[0])

def getAreasUpTo(map_in, size):
    areas = []
    tmp = os.popen("r.stats -lnNc %s | sort -n -k2 | awk %s" % (map_in, AWK_cmd_3(size))).readlines()
    for x in tmp:
        areas.append(int(x))
    return areas

try:
    from subprocess import DEVNULL # Python 3.
except ImportError:
    #DEVNULL = open('trace_step3', 'wb')
    DEVNULL = open(os.devnull, 'wb')
'''
MAIN
'''
os.environ['GRASS_MESSAGE_FORMAT'] = 'silent'

timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
print('---------- HRU-delin Step 3 started ---------------------------------------------', timestr)
parms = ConfigParser.ConfigParser(allow_no_value=True)
parms_file = "hrudelin_config.cfg"
parms.read(parms_file)
directory_out = parms.get("dir_out", "files")
min_area = int(parms.get("hrus_min_surface", "surface"))

# Import rasters in grass envir
print('---------- HRU-delin Step 3 : Importing the rasters in GRASS')

dem_recl = directory_out + '/' + 'step1_dem_reclass.tif'
grass.run_command('g.proj', flags='p', georef=dem_recl, stdout = DEVNULL, stderr = DEVNULL)
    
data_list = []
mask_list = ""
if parms.get('layer_overlay', 'dem') == 'x':
	dem = 'dem'
	grass.run_command('r.in.gdal', flags='o', input=dem_recl, output=dem, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
	data_list.append("dem")
	mask_list += "dem_msk,"

if parms.get('layer_overlay', 'slope') == 'x':
	slp_recl = directory_out + '/' + 'step1_slope_reclass.tif'  
	grass.run_command('r.in.gdal', flags='o', input=slp_recl, output='slp', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
	data_list.append("slp")
	mask_list += "slp_msk,"

if parms.get('layer_overlay', 'aspect') == 'x':
	asp_recl = directory_out + '/' + 'step1_aspect_reclass.tif'
	grass.run_command('r.in.gdal', flags='o', input=asp_recl, output='asp', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
	data_list.append("asp")
	mask_list += "asp_msk,"

subbasins = directory_out + '/' + 'step1_subbasins.tif'
grass.run_command('r.in.gdal', flags='o', input=subbasins, output='subbasins', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
data_list.append("subbasins")
mask_list += "subbasins_msk,"

watersheds = 'watersheds'
basins = directory_out + '/' + 'step2_watersheds.tif'
grass.run_command('r.in.gdal', flags='o', input=basins, output='watersheds', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
data_list.append("watersheds")
mask_list += "watersheds_msk,"

for data in parms.items('data'):
	name = data[0]
	data_file = (directory_out + '/' + 'step1_' + data[1])
	grass.run_command('r.in.gdal', flags='o'
					, input=data_file
                    , output=name, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
	data_list.append(name)
	mask_list += name + "_msk,"

mask_list = mask_list.strip(",")

grass.run_command('g.region', rast='watersheds', stdout = DEVNULL, stderr = DEVNULL)

# Import mask(= all watersheds)
mask_in = directory_out + '/' + 'step2_mask.tif'
mask_wk = 'mask_wk'
grass.run_command('r.in.gdal', flags='o', input=mask_in, output=mask_wk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.null', map=mask_wk, null=0, setnull=1, stdout = DEVNULL, stderr = DEVNULL)
# Cut out every layer with mask
for layer in data_list:
    grass.run_command('r.patch', input='mask_wk,%s' %(layer), output='%s_msk' %(layer), overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
    grass.run_command('r.null', map='%s_msk' % layer, setnull=0, stdout = DEVNULL, stderr = DEVNULL)

# Create a first raster of hrus = combination (overlay) of layers
timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
print('---------- HRU-delin Step 3 : Creating the cross product of the layers', timestr) 
grass.run_command('r.cross', flags='z', input='%s' %(mask_list), 
                   output='hrus_tmp', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.out.gdal', input='hrus_tmp', 
                   output=directory_out + '/step3_hrutmp.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Attrib a unique cat to each HRU generated
clumps = 'clumps'
grass.run_command('r.clump', input='hrus_tmp', output=clumps, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.out.gdal', input=clumps, 
                   output=directory_out + '/step3_clumps.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
# Start merging
grass.run_command('g.region', rast=clumps, stdout = DEVNULL, stderr = DEVNULL)
buffer_distance = int(math.ceil(float(os.popen("g.region -m").readlines()[4].split("=")[1])))
buffer_distance_ew = int(math.ceil(float(os.popen("g.region -m").readlines()[5].split("=")[1])))

if (abs(buffer_distance) != abs(buffer_distance_ew)):
		max_buffer = max(buffer_distance, buffer_distance_ew)+1
		buffer_distance = max_buffer

grass.run_command('r.mask', input=clumps, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
basin_ids = os.popen("r.stats -nN subbasins").readlines()

clump_cp = 'clumps_cp'
grass.run_command('r.mapcalc', clumps_cp=clumps, stdout = DEVNULL, stderr = DEVNULL)

result = 'result'
grass.run_command('r.mapcalc', result='null()', stdout = DEVNULL, stderr = DEVNULL)
timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
print('---------- HRU-delin Step 3 : Starting HRUs construction, may take some time ... ', timestr)

# Subbasin loop
i = 1 
for id in basin_ids:
        os.system("r.mask -o subbasins_msk maskcats=%s" % id.rstrip("\n")) 
        grass.run_command('g.region', rast=clumps, stdout = DEVNULL, stderr = DEVNULL)
        grass.run_command('g.region', zoom='MASK', stdout = DEVNULL, stderr = DEVNULL)
        # Get size of subbasin
        #subbasin_size = os.popen("r.stats -Nnc MASK").read()
        subbasin_size = int(os.popen("r.stats -Nnc MASK").read().split(" ")[1].rstrip("\n"))
        if (subbasin_size <= min_area):
            os.system("r.statistics --o base=MASK cover=clumps method=max output=hruid")
            hruid = int((os.popen("r.stats -lnNc hruid").read().split(" ")[1]))
            os.system("r.mapcalc newmap3='if(MASK,%d,null())'" % hruid)
        else:
           # LOOP THAT ELIMINATES PIXELS ONLY
           initmap = "clumps"
           counter_old = -1
           while (True):
               counter = count_only(initmap, 1)
               if (counter != counter_old):
                   reclass(initmap, "clumps_sz1", 1)
                   counter_old = counter
                   os.system("r.mapcalc base=clumps_sz1")
                   # MERGE HORIZONTAL/VERTICAL ADJACENT PIXELS
                   os.system("r.mapcalc map_a='if((not(isnull(base))&&not(isnull(base[0,1]))),base[0,1],if((not(isnull(base))&&not(isnull(base[1,0]))),base[1,0],base))'")
                   os.system("r.patch --o input=map_a,clumps_cp output=newmap")
                   os.system("r.mapcalc clumps_cp=newmap")
                   initmap = "newmap"
               else:
                   break

           # START TO ELIMINATE REMAINING AND ISOLATED PIXELS
           os.system("r.mapcalc newmap_sz1=clumps_sz1")
           os.system("r.buffer --o input=newmap_sz1 output=buffer distance=%d" % buffer_distance)
           os.system("r.mapcalc newmap2='if((buffer==2),newmap,null())'")
           os.system("r.stats -cnN newmap | sort -k 1b,1 > tmp/out1")
           os.system("r.stats -nN newmap2 | sort -k 1b,1 > tmp/out2")
           os.system("join tmp/out1 tmp/out2 | awk '{print $1 \" = \" $1 \" \" $2}' > tmp/out3")
           os.system("cat tmp/out3 | r.reclass --o input=newmap output=test")
           os.system("g.region rast=clumps")
           os.system("r.buffer --o input=MASK output=buf_mask distance=%d" % buffer_distance)
           os.system("r.mask -r MASK")
           os.system("g.region zoom=buf_mask")
           os.system("echo '2 = 2 9999999' | r.reclass --o input=buf_mask output=buf_mask_new")
           os.system("r.patch --o input=buf_mask_new,test output=sum_buf")
           os.system("r.mapcalc newmap3='if((isnull(newmap_sz1)),newmap,eval(a=@sum_buf[0,-1],b=@sum_buf[-1,0],c=@sum_buf[0,1],d=@sum_buf[1,0],m=min(a,b,c,d),if((m==int(a)),sum_buf[0,-1],if((m==int(b)),sum_buf[-1,0],if((m==int(c)),sum_buf[0,1],if((m==int(d)),sum_buf[1,0]))))))'")

           # START ELIMINATING NOW FOR AREAS GREATER THAN 1 PIXEL
           for sz in range (2, min_area + 1):
               if sz not in getAreasUpTo("newmap3", min_area):
                   pass
               else:
                   initmap = "newmap3"
                   counter_old = -1
                   while (True):
                       counter = count_only(initmap, sz)
                       if (counter != counter_old):
                           reclass(initmap, "clumps_sz2", sz)
                           counter_old = counter
                           os.system("r.mapcalc base=clumps_sz2")
                           os.system("r.null map=base null=0")
                           os.system("r.mapcalc map_d='if((not(isnull(base))),eval(a=base[0,-1],b=base[-1,0],c=base[0,1],d=base[1,0],i=base,max(a,b,c,d,i)))'")
                           os.system("r.statistics --o base=base cover=map_d method=mode output=map1")
                           os.system("r.stats -lnN map1 | tail -n +2 | awk '{print $1 \" = \" $2}' | r.reclass --o input=base output=base_new")
                           os.system("r.patch --o input=base_new,newmap3 output=newout")
                           os.system("r.mapcalc newmap3=newout")
                           initmap = "newmap3"
                       else:
                           # ELIMINATION STOPPED
                           break
                   os.system("r.mask -o subbasins maskcats=%s" % id.rstrip("\n"))
                   os.system("r.stats -cnN newmap3 | awk '{print $1 \" = \" $1 \" \" $2}' | r.reclass --o input=newmap3 output=newout_sz")
                   reclass("newmap3", "newmap_sz2", sz)
                   os.system("r.buffer --o input=newmap_sz2 output=buffer distance=%d" % buffer_distance)
                   os.system("echo '1 thru 2 = 1' | r.reclass --o buffer output=buffer_new")
                   os.system("r.mapcalc b1='if((buffer_new==1&&isnull(newmap_sz2)),newmap_sz2[-1,0],null())'")
                   os.system("r.mapcalc b2='if((buffer_new==1&&isnull(newmap_sz2)),newmap_sz2[1,0],null())'")
                   os.system("r.mapcalc b3='if((buffer_new==1&&isnull(newmap_sz2)),newmap_sz2[0,-1],null())'")
                   os.system("r.mapcalc b4='if((buffer_new==1&&isnull(newmap_sz2)),newmap_sz2[0,1],null())'")
                   os.system("r.patch --o input=b1,b2,b3,b4 output=b")
                   os.system("r.mapcalc buffer1='if((buffer==2),int(@newout_sz),null())'")
                   os.system("r.statistics --o base=b cover=buffer1 method=min output=buffer_min")
                   os.system("r.mapcalc buffer3='if((buffer1==@buffer_min),newmap3,null())'")
                   os.system("r.statistics --o base=buffer_min cover=buffer3 method=max output=final")
                   os.system("r.stats -lnN final | awk '{print $1 \" = \" $2}' | r.reclass --o input=newmap_sz2 output=newmap_sz2_rec")
                   os.system("r.patch --o input=newmap_sz2_rec,newmap3 output=newout2")
                   os.system("r.mapcalc newmap3=newout2")

            # SOLVE STRANGE PROBLEM OF EDGED WATERSHEDS IF EXISTS
           os.system("r.mapcalc map_a='if((not(isnull(newmap3))&&isnull(newmap3[-1,0])&&isnull(newmap3[1,0])&&isnull(newmap3[0,-1])&&isnull(newmap3[0,1])),-1,newmap3)'")
           check = int(os.popen("r.stats -lnN map_a").readlines()[0])
           if (check != -1):
               # NOTHING TO DO
               pass
           else:
               os.system("r.mapcalc map_b='if((map_a==-1),if((map_a[-1,1]>-1|||map_a[1,1]>-1|||map_a[1,-1]>-1|||map_a[-1,-1]>-1),-2,null()),map_a)'")
               os.system("r.mapcalc map_c='if((map_b==-2),if((not(isnull(map_b[-1,1]))),map_b[-1,1],if((not(isnull(map_b[1,1]))),map_b[1,1],if((not(isnull(map_b[1,-1]))),map_b[1,-1],if((not(isnull(map_b[-1,-1]))),map_b[-1,-1])))),map_b)'")
               os.system("r.mapcalc map_d='if((map_a==-1),1,null())'")

               os.system("r.buffer --o input=map_d output=map_d2 distances=%d" % buffer_distance)
               os.system("echo '1 thru 2 = 1' | r.reclass --o map_d2 output=map_d3")
               os.system("r.mask -r MASK")
               os.system("r.clump --o input=map_d3 output=map_d4")
               os.system("r.mapcalc map_d5='if((not(isnull(map_d4[-1,0]))&&not(isnull(map_d4[1,0]))&&not(isnull(map_d4[0,-1]))&&not(isnull(map_d4[0,1]))),map_d4,null())'")

               ##os.system("r.clump --o input=map_d output=map_e")
               os.system("r.statistics --o base=map_d5 cover=map_c method=max output=map_f") #map_e
               os.system("r.mapcalc map_g='int(@map_f)'")
               os.system("r.patch --o input=map_g,newmap3 output=map_h")
               os.system("r.mapcalc newmap3=map_h")

#        try:
        os.system("r.mask -r MASK")
#        except:
#            pass
            
        os.system("g.region rast=clumps")
        os.system("r.patch --o input=result,newmap3 output=result_tmp")
        os.system("r.mapcalc result=result_tmp")

        os.system("r.mapcalc newmap3='null()'")
        os.system("r.mapcalc clumps_cp=clumps")

        i += 1
 
timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
print('---------- HRU-delin Step 3 : Reassign category values of HRUs', timestr)
os.system("r.stats -nN result | awk 'BEGIN{i=1} {print $1 \"=\" i;i++} END{}' > tmp/result_rules")
os.system("r.reclass --o input=result output=result_rcl rules='tmp/result_rules'")
os.system("r.mapcalc result_rcl2='if((result_rcl==0),null(),result_rcl)'")
os.system("r.colors result_rcl2 color=random")
hrus = 'result_rcl2'

# Save rasters 
grass.run_command('r.out.gdal', input=hrus, output=directory_out + '/' + 'step3_hrus.tif', overwrite='True', stdout=DEVNULL, stderr=DEVNULL)
timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
print('---------- HRU-delin Step 3 ended   ---------------------------------------------', timestr)

