# -*- coding: utf-8 -*-
'''

 MODULE:       hru-delin_parms_J2000

 AUTHOR(S):    adapted from GRASS-HRU (ILMS) - JENA University
               IRSTEA - Christine Barachet 
               

 PURPOSE:      calculates the topology
 			   generates parameters files for J2000

'''

import string, os, sys, time, types
import zipfile
import ConfigParser
import grass.script as grass
import numpy as np

from osgeo import gdal
from gdalconst import *
from osgeo import ogr

# Format table N:M
AWK_cmd_1 = "modules/awk/format_nm.awk"

# Find max rate in N:M (+ associated id)
AWK_cmd_2 = "modules/awk/find_max.awk"

# Script to use if no cycle dissolving is chosen
AWK_cmd_3 = "modules/awk/find_max_no_cycle_opt.awk"

# Transform back to original HRU*TO_REACH*TO_PAR format
# in case of cycle removing
AWK_cmd_4 = "modules/awk/orig_format.awk"

# Replace empty headerline by new header
AWK_cmd_5 = "modules/awk/repl_header.awk"

'''
 MAIN
'''
try:
    from subprocess import DEVNULL # Python 3.
except ImportError:
    DEVNULL = open(os.devnull, 'wb')
    #DEVNULL = open('trace_step4', 'wb')
os.environ['GRASS_MESSAGE_FORMAT'] = 'silent'

print('------------- HRU-delin parms_J2000 Started -------------')
parms = ConfigParser.ConfigParser(allow_no_value=True)
parms_file = "hrudelin_config.cfg"
parms.read(parms_file)
directory_out = parms.get("dir_out", "files")
dir_results = parms.get("dir_out", "results")

# Import rasters in grass envir
hrus ='hrus'
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step3_hrus.tif', 
                  output=hrus, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
dem = 'dem'
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step1_dem_filled.tif', 
                  output=dem, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
slope = 'slope'
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step1_slope.tif',
                  output=slope, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
aspect = 'aspect' 
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step1_aspect.tif' ,
                  output=aspect, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
aspect_rcl = 'aspect_rcl' 
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step1_aspect_reclass.tif' , 
                  output=aspect_rcl, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
drain = 'drain'
grass.run_command('r.in.gdal', flags='o',
                  input=directory_out + '/step1_drain.tif',
                  output=drain, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
accum = 'accum'
grass.run_command('r.in.gdal', flags='o',
                  input=directory_out + '/step1_accum.tif',
                  output=accum, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
subbasins = 'subbasins'
grass.run_command('r.in.gdal', flags='o',
                  input=directory_out + '/step1_subbasins.tif',
                  output=subbasins, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)
watersheds = 'watersheds' 
grass.run_command('r.in.gdal', flags='o', 
                  input=directory_out + '/step2_watersheds.tif',
                  output=watersheds, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)

for data in parms.items('data'):
	name = data[0]
	data_file = (directory_out + '/step1_' + data[1])
	grass.run_command('r.in.gdal', flags='o', 
                       input=data_file,
                       output=name, overwrite='True',
                       stdout = DEVNULL, stderr = DEVNULL)

if str(parms.get("irrigation", "irrig_rast")) != '':
    irrigation = 'irrigation'
    grass.run_command('r.in.gdal', flags='o', 
                       input=directory_out + '/step1_irrigation.tif',
                       output=irrigation, overwrite='True',
                       stdout = DEVNULL, stderr = DEVNULL)
    grass.run_command('r.null', map=irrigation, null=0, stdout = DEVNULL, stderr = DEVNULL)
    
# Import mask(= all watersheds)
mask_in = directory_out + '/' + 'step2_mask.tif'
mask_wk = 'mask_wk'
grass.run_command('r.in.gdal', flags='o', input=mask_in, output=mask_wk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.null', map=mask_wk, null=0, setnull=1, stdout = DEVNULL, stderr = DEVNULL)

grass.run_command('g.region', rast=subbasins, stdout = DEVNULL, stderr = DEVNULL)

# Set mask to all HRUs
##optim
##grass.run_command('r.mask', input=hrus, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Prepare for topology construction
streams_wk = 'streams_wk'
grass.run_command('r.in.gdal', flags='o',
                  input=directory_out + '/step2_streams_new.tif',
                  output=streams_wk, overwrite='True',
                  stdout = DEVNULL, stderr = DEVNULL)

# Save streams raster as shapefile
streams_thinned = 'streams_thinned'
grass.run_command('r.thin', input=streams_wk, output=streams_thinned, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)    
streams_shape = dir_results + '/streams_shape' 
grass.run_command('r.to.vect', flags='v', input=streams_thinned, output='streams_shp', feature='line', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('v.out.ogr', flags='c', input='streams_shp', type='line', dsn=streams_shape, stdout = DEVNULL, stderr = DEVNULL)
# Set null in stream and negate
os.system("r.null streams_wk setnull=0")
os.system("r.mapcalc str_wk='if((streams_wk>0),-streams_wk,streams_wk)'")

# Subbasins loop
print('------------- subbasin loop Started -------------', time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime()))
subbasins_msk = 'subbasins_msk'
grass.run_command('r.patch', input='mask_wk,subbasins', output=subbasins_msk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.null', map=subbasins_msk, setnull=0, stdout = DEVNULL, stderr = DEVNULL)
basin_ids = os.popen("r.stats -Nn subbasins_msk").readlines()
i = 1
for id in basin_ids:
    os.system("r.mask -o input=subbasins maskcats=%s" % id.rstrip("\n"))
    os.system("g.region rast=MASK")

    # OVERLAY HRUS WITH STREAM MAP
    os.system("r.patch --o -z input=str_wk,hrus output=result_cp")

	# HRU BORDERS
    os.system("r.mapcalc hru_border='if((result_cp==result_cp[-1,-1]&&&not(isnull(result_cp[-1,-1]))&&&result_cp==result_cp[-1,0]&&&not(isnull(result_cp[-1,0]))&&&result_cp==result_cp[-1,1]&&&not(isnull(result_cp[-1,1]))&&&result_cp==result_cp[0,1]&&&not(isnull(result_cp[0,1]))&&&result_cp==result_cp[1,1]&&&not(isnull(result_cp[1,1]))&&&result_cp==result_cp[1,0]&&&not(isnull(result_cp[1,0]))&&&result_cp==result_cp[1,-1]&&&not(isnull(result_cp[1,-1]))&&&result_cp==result_cp[0,-1]&&&not(isnull(result_cp[0,-1]))),null(),result_cp)'")

	# NEIGHBOUR HRU ON BORDERS
    os.system("r.mapcalc hru_nb='if((drain==2),if((hru_border==hru_border[-1,0]),null(),hru_border[-1,0]),if((drain==3),if((hru_border==hru_border[-1,-1]),null(),hru_border[-1,-1]),if((drain==4),if((hru_border==hru_border[0,-1]),null(),hru_border[0,-1]),if((drain==5),if((hru_border==hru_border[1,-1]),null(),hru_border[1,-1]),if((drain==6),if((hru_border==hru_border[1,0]),null(),hru_border[1,0]),if((drain==7),if((hru_border==hru_border[1,1]),null(),hru_border[1,1]),if((drain==8),if((hru_border==hru_border[0,1]),null(),hru_border[0,1]),if((drain==1),if((hru_border==hru_border[-1,1]),null(),hru_border[-1,1])))))))))'")

	# GET HRU AREAS
    os.system("r.stats -cnN in=hrus | awk '{print $1 \"=\" $1\" \"$2'} | r.reclass --o input=hrus output=hru_sz")
    # CREATE OVERLAY MAP FOR FURTHER PROCESSING
    os.system("r.cross --o -z input=hru_border,hru_nb,hru_sz output=res1")
	# FLOW ACCUMULATION ANALYSIS ON res1
    os.system("r.mapcalc acc_int='int(accum)'")
    os.system("r.statistics --o base=res1 cover=acc_int method=sum output=sum_map")
    os.system("r.mapcalc acc_sum=sum_map")
    os.system("r.null acc_sum setnull=0")
   	
	# NEW FOR MAX ACCUMULATION (FOR BJOERN CIRCLE ELIM)
    os.system("r.statistics --o base=hrus cover=acc_int method=max output=max_map")
    os.system("r.mapcalc acc_max=max_map")
    os.system("r.null acc_max setnull=0")
	# END NEW

	# AWK PROCESSING
	# acc_max ADDED TO r.stats
    ## subb100
    new_id = str(int(id.rstrip("\n")) * 100)
    os.system("r.stats -lnN res1,acc_sum,acc_max | awk -v BASIN=%s -f %s" % (new_id, AWK_cmd_1))
    i += 1

print('------------- subbasin loop Ended -------------', time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime()))

# REMOVE EMPTY HEADER LINE
os.system("awk -v MODE=%i -f %s topologie_to_hru.par_tmp > topologie_to_hru.par_tmp2" % (1, AWK_cmd_5))
os.system("awk -v MODE=%i -f %s topologie_bfl.par_tmp > topologie_bfl.par_tmp2" % (2, AWK_cmd_5))
              
# ...AND SORT
os.system("(head -n 2 topologie_to_hru.par_tmp2 && tail -n +3 topologie_to_hru.par_tmp2 | sort -n -k 1) > topologie_to_hru.par")
os.system("(head -n 2 topologie_bfl.par_tmp2 && tail -n +3 topologie_bfl.par_tmp2 | sort -n -k 1) > topologie_bfl.par")

# Find maximum destination rate (need to be updated because of additiona 2 line header
# output adapted to requirements for Bjoern tool (see next)

if (parms.get("topology", "dissolve_cycle")) == 'y':
	print('------------- Dissolve cycles -------------')		
	# adaptation N1
	os.system("awk -f %s topologie_bfl.par > topologie_n1_tmp.par" % AWK_cmd_2)

	# ADAPTION NM, E.G. ELIMINATION TOOL REQUIRES ,'#', AS SEPARATOR
	#os.system("(head -n 2 topologie_to_hru.par && tail -n +3 topologie_to_hru.par | sed \"s/\t/,'#',/g\") > topologie_to_hru_tmp.par")
	#os.system("cp topologie_to_hru_tmp.par topologie_to_hru.par")
	#os.system("(head -n 2 topologie_bfl.par && tail -n +3 topologie_bfl.par | sed \"s/\t/,'#',/g\") > topologie_bfl_tmp.par")
	#os.system("cp topologie_bfl_tmp.par topologie_bfl.par")
			
	# N1: REMOVE CYCLES USING BJOERN TOOL (Zirkelkill_1zuN.jar, NOT UPDATED)
	# RESULTS IN topologie_n1.par FILE
	os.system("/usr/bin/java -jar /usr/share/java/Zirkelkill_1zuN.jar topologie_n1_tmp.par topologie_n1_tmp2")

	# NEED ORIGINAL TABLE FORMAT (AND ADD HEADER) - E.G. FOR FINAL QGIS HRU BROWSER
	os.system("awk -f %s topologie_n1_tmp2.par > topologie_n1.par" % AWK_cmd_4)
else:
	os.system("sed \"1,2d\" topologie_to_hru.par > topologie_to_hru_nohead.par")
	os.system("cp topologie_to_hru_nohead.par topologie_to_hru.par")
	os.system("sed \"1,2d\" topologie_bfl.par > topologie_bfl_nohead.par")
	os.system("cp topologie_bfl_nohead.par topologie_bfl.par")
	os.system("awk -f %s topologie_bfl.par > topologie_n1.par" % AWK_cmd_3)

# CREATE REACH ROUTING FIRST
os.system("r.mask -r MASK")

# Call the module that creates the file reach.par
reach_par = dir_results +'/reach.par'
os.system("r.reach.par streams=streams_wk watershed=hrus drainage=drain dem=dem output=%s" % reach_par)

# Convert hru raster to shape polygon
grass.run_command('g.region', rast=hrus)
grass.run_command('g.region', zoom=hrus)
grass.run_command('r.to.vect', input=hrus, output='hrus_v', feature='area', overwrite='True')

grass.run_command('db.droptable', flags='f', table='hrus_v_cat')
grass.run_command('r.to.vect', flags='v', input=hrus, output='hrus_v_cat', feature='area', overwrite='True')

grass.run_command('v.extract', input='hrus_v', type='centroid', output='hrus_c', overwrite='True')
grass.run_command('v.type', input='hrus_c', output='hrus_c_pnt', type='centroid,point', overwrite='True')

# Create parameters file for J2000

#  header
ranges = {}
ranges["ID"] = [0,999999,"n/a"]

f = open("tmp/header","w+")
timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
header = "# hru.par created " + timestr + " by HRU-DELIN_BATCH\n"
header += "ID"
header_min_val = str(ranges["ID"][0])
header_max_val = str(ranges["ID"][1])
header_unit = ranges["ID"][2]
sql = "cat"

nres = int(float(os.popen("g.region -m").readlines()[4].split("=")[1]))
sres = int(float(os.popen("g.region -m").readlines()[5].split("=")[1]))
single_pixel_size = nres*sres

os.system("r.stats -cnN input=hrus | awk '{print $1\"=\"$2*%d}' | r.reclass --o input=hrus output='hrus_sz'" % single_pixel_size)
    
os.system("echo 'ALTER TABLE hrus_c_pnt ADD area INTEGER' | db.execute")
grass.run_command('v.what.rast', vect='hrus_c_pnt', rast='hrus_sz', col='area')
header += "\tarea"
ranges["area"] = [0,9999999,"m2"]
header_min_val += "\t"+str(ranges["area"][0])
header_max_val += "\t"+str(ranges["area"][1])
header_unit += "\t"+ranges["area"][2]
sql += ",area"     

if (parms.get("layer_overlay", "dem")) == 'x':
    dem_avg='dem_avg'
    dem_int='dem_int'
    grass.run_command('r.mapcalc', dem_int='int(dem)')
    grass.run_command('r.statistics', base=hrus, cover=dem_int, method='average', output=dem_avg, overwrite='True')
    os.system("r.mapcalc dem_avg_cval=@dem_avg")
    os.system("echo 'ALTER TABLE hrus_c_pnt ADD elevation INTEGER' | db.execute")
    grass.run_command('v.what.rast', vect='hrus_c_pnt', rast='dem_avg_cval', col='elevation', overwrite='True')
    header += "\televation"
    ranges["elevation"] = [0,10000,"m"]
    header_min_val += "\t"+str(ranges["elevation"][0])
    header_max_val += "\t"+str(ranges["elevation"][1])
    header_unit += "\t"+ranges["elevation"][2]
    sql += ",elevation"

if (parms.get("layer_overlay", "slope")) == 'x':
    slope_mult = 'slope_mult'
    slope_labeled='slope_labeled'
    slope_avg = 'slope_avg'
    slope_avg_cval = 'slope_avg_cval'
    grass.run_command('r.mapcalc', slope_mult='int(slope*1000.0)')
    os.system("r.stats -nN input=slope_mult | awk '{print $1\"=\"$1\" \"$1/1000}' | r.reclass --o input=slope_mult output=slope_labeled")
    
    grass.run_command('r.statistics', flags='c',base=hrus, cover=slope_labeled, method='average', output=slope_avg, overwrite='True')
    grass.run_command('r.mapcalc', slope_avg_cval='int(@slope_avg*1000)/1000.0')
    os.system("echo 'ALTER TABLE hrus_c_pnt ADD slope DOUBLE PRECISION' | db.execute")
    os.system("v.what.rast vect=hrus_c_pnt rast=slope_avg_cval col=slope")
    header += "\tslope"
    ranges["slope"] = [0,90,"deg"]
    header_min_val += "\t"+str(ranges["slope"][0])
    header_max_val += "\t"+str(ranges["slope"][1])
    header_unit += "\t"+ranges["slope"][2]
    sql += ",slope"

if (parms.get("layer_overlay", "aspect")) == 'x':
    os.system("r.statistics --o base=hrus cover=aspect_rcl method=mode output=aspect_md")
    grass.run_command('r.mapcalc', aspect_mj='if((aspect_rcl==@aspect_md),aspect,null())')
    grass.run_command('r.mapcalc', aspect_mjj='if((aspect_mj>315.0),aspect_mj-360.0,aspect_mj)')
    grass.run_command('r.mapcalc', aspect_int='int(aspect_mjj)')
    os.system("r.statistics --o base=hrus cover=aspect_int output=aspect_avg method=average")
    os.system("r.mapcalc aspect_avg_cval=@aspect_avg")
    os.system("echo 'ALTER TABLE hrus_c_pnt ADD aspect INTEGER' | db.execute")
    os.system("v.what.rast vect=hrus_c_pnt rast=aspect_avg_cval col=aspect")
    header += "\taspect"
    ranges["aspect"] = [0,360,"deg"]
    header_min_val += "\t"+str(ranges["aspect"][0])
    header_max_val += "\t"+str(ranges["aspect"][1])
    header_unit += "\t"+ranges["aspect"][2]
    sql += ",aspect"

os.system("echo 'ALTER TABLE hrus_c_pnt ADD x DOUBLE PRECISION' | db.execute")
os.system("echo 'ALTER TABLE hrus_c_pnt ADD y DOUBLE PRECISION' | db.execute")
os.system("v.to.db map=hrus_c_pnt option=coor column=x,y")
header += "\tx\ty"
ranges["centroid"] = [0,9999999,"m"]
header_min_val += "\t"+str(ranges["centroid"][0])+"\t"+str(ranges["centroid"][0])
header_max_val += "\t"+str(ranges["centroid"][1])+"\t"+str(ranges["centroid"][1])
header_unit += "\t"+ranges["centroid"][2]+"\t"+ranges["centroid"][2]
sql += ",x,y"

os.system("echo 'ALTER TABLE hrus_c_pnt ADD watershed INTEGER' | db.execute")
os.system("v.what.rast vect=hrus_c_pnt rast=watersheds col=watershed")
header += "\twatershed"
ranges["watershed"] = [0,999999,"n/a"]
header_min_val += "\t"+str(ranges["watershed"][0])
header_max_val += "\t"+str(ranges["watershed"][1])
header_unit += "\t"+ranges["watershed"][2]
sql += ",watershed"


os.system("echo 'ALTER TABLE hrus_c_pnt ADD subbasin INTEGER' | db.execute")
os.system("v.what.rast vect=hrus_c_pnt rast=subbasins col=subbasin")
header += "\tsubbasin"
ranges["subbasin"] = [0,999999,"n/a"]
header_min_val += "\t"+str(ranges["subbasin"][0])
header_max_val += "\t"+str(ranges["subbasin"][1])
header_unit += "\t"+ranges["subbasin"][2]
sql += ",subbasin"

data_tmp2 = 'data_tmp2'
data_md = 'data_md'
data_cval = 'data_cval'
for data in parms.items('data'):
    col_name = (data[0] + 'ID')
    os.system("r.mapcalc data_tmp2='int(%s)'" % data[0])
    grass.run_command('r.statistics', base=hrus, cover=data_tmp2, method='mode', output=data_md, overwrite='True')
    os.system("r.mapcalc data_cval=@data_md")
    os.system("echo 'ALTER TABLE hrus_c_pnt ADD %s INTEGER' | db.execute" % col_name)
    os.system("v.what.rast vect=hrus_c_pnt rast=data_cval col=%s" % col_name)
    header += "\t" + col_name
    ranges["data1"] = [0,9999,"n/a"]
    header_min_val += "\t"+str(ranges["data1"][0])
    header_max_val += "\t"+str(ranges["data1"][1])
    header_unit += "\t"+ranges["data1"][2]
    sql += "," + col_name

# FROM N1 ROUTING FILE, JUST SKIP FIRST TWO LINES (HEADER)
os.system("tail -n +3 %s | awk 'BEGIN{}{print $1\"=\"$2 >> \"tmp/to_poly\";print $1\"=\"$3*-1 >> \"tmp/to_reach\";}END{}'" % 'topologie_n1.par')

os.system("cat tmp/to_poly | r.reclass --o input=hrus output=to_poly_map")
os.system("cat tmp/to_reach | r.reclass --o input=hrus output=to_reach_map")
os.system("echo 'ALTER TABLE hrus_c_pnt ADD to_poly INTEGER' | db.execute")
os.system("echo 'ALTER TABLE hrus_c_pnt ADD to_reach INTEGER' | db.execute")
os.system("v.what.rast vect=hrus_c_pnt rast=to_poly_map col=to_poly")
os.system("v.what.rast vect=hrus_c_pnt rast=to_reach_map col=to_reach")
header += "\tto_poly"
sql += ",to_poly"
header += "\tto_reach"
sql += ",to_reach"
ranges["topo"] = [0,999999,"n/a"]
header_min_val += "\t"+str(ranges["topo"][0])+"\t"+str(ranges["topo"][0])
header_max_val += "\t"+str(ranges["topo"][1])+"\t"+str(ranges["topo"][1])
header_unit += "\t"+ranges["topo"][2]+"\t"+ranges["topo"][2]

if str(parms.get("irrigation", "irrig_rast")) != '':
    col_name = ('irrigation')
    irrig_md = 'irrig_md'
    os.system("r.mapcalc data_tmp2='int(%s)'" % irrigation)
    grass.run_command('r.statistics',flags='c', base=hrus, cover=data_tmp2, method='mode', output=irrig_md, overwrite='True')
    os.system("r.mapcalc irrig_cval=@irrig_md")
    os.system("echo 'ALTER TABLE hrus_c_pnt ADD %s INTEGER' | db.execute" % col_name)
    os.system("v.what.rast vect=hrus_c_pnt rast=irrig_cval col=%s" % col_name)
    header += "\t" + col_name
    ranges["irrigation"] = [0,1,"n/a"]
    header_min_val += "\t"+str(ranges["irrigation"][0])
    header_max_val += "\t"+str(ranges["irrigation"][1])
    header_unit += "\t"+ranges["irrigation"][2]
    sql += ",irrigation"

os.system("echo 'UPDATE hrus_c_pnt SET cat=value' | db.execute")

# Save files 
grass.run_command('v.db.connect', flags='o', table='hrus_c_pnt', map='hrus_v_cat', key='cat', overwrite='True')
hrus_shp = dir_results +'/hrus.shp'
grass.run_command('v.out.ogr', flags='c', input='hrus_v_cat', type='area', dsn=hrus_shp)

##optim
vals_file='vals_file'
os.system("db.select -c 'hrus_c_pnt' sql='SELECT %s from %s' | sort -n -u > tmp/vals_file" % (sql, 'hrus_c_pnt'))
##optim

# Concatenate header and data
f.write(header+"\n")
f.write(header_min_val+"\n")
f.write(header_max_val+"\n")
f.write(header_unit+"\n")
f.close()
footer = "# end of hru.par"

par = dir_results + '/hru.par'
os.system("echo '%s' | cat tmp/header tmp/vals_file - | tr '|' '\t' > %s" % (footer, par))

os.system("mv topologie_* tmp")
save_config = directory_out + '/hrudelin_config_save.cfg'
os.system("cp hrudelin_config.cfg '%s' " % save_config)
print('------------- HRU-delin parms_J2000 Ended -------------')
